@@ -3,7 +3,7 @@
 //!
 //! See `Manifest::from_slice`.
 
-
+use std::fmt::Display;
 use std::fs;
 use std::io;
 use std::path::Path;
@@ -106,6 +106,9 @@ pub struct Workspace<Metadata = Value> {
     pub exclude: Vec<String>,
 
     pub metadata: Option<Metadata>,
+
+    #[serde(skip_serializing_if = "Option::is_none")]
+    pub resolver: Option<Resolver>,
 }
 
 fn default_true() -> bool {
@@ -120,6 +123,7 @@ impl Manifest<Value> {
     /// Parse contents of a `Cargo.toml` file already loaded as a byte slice.
     ///
     /// It does not call `complete_from_path`, so may be missing implicit data.
+    #[inline(always)]
     pub fn from_slice(cargo_toml_content: &[u8]) -> Result<Self, Error> {
         Self::from_slice_with_metadata(cargo_toml_content)
     }
@@ -127,6 +131,7 @@ impl Manifest<Value> {
     /// Parse contents from a `Cargo.toml` file on disk.
     ///
     /// Calls `complete_from_path`.
+    #[inline]
     pub fn from_path(cargo_toml_path: impl AsRef<Path>) -> Result<Self, Error> {
         Self::from_path_with_metadata(cargo_toml_path)
     }
@@ -136,6 +141,7 @@ impl Manifest<Value> {
     /// Note: this is **not** a file name, but file's content. See `from_path`.
     ///
     /// It does not call `complete_from_path`, so may be missing implicit data.
+    #[inline(always)]
     pub fn from_str(cargo_toml_content: &str) -> Result<Self, Error> {
         Self::from_slice_with_metadata(cargo_toml_content.as_bytes())
     }
@@ -192,7 +198,12 @@ impl<Metadata: for<'a> Deserialize<'a>> Manifest<Metadata> {
         fs: impl AbstractFilesystem,
     ) -> Result<(), Error> {
         if let Some(ref package) = self.package {
-            let src = fs.file_names_in("src").unwrap_or_default();
+            let src = match fs.file_names_in("src") {
+                Ok(src) => src,
+                Err(err) if err.kind() == io::ErrorKind::NotFound => Default::default(),
+                Err(err) => return Err(err.into()),
+            };
+
             if let Some(ref mut lib) = self.lib {
                 lib.required_features.clear(); // not applicable
             }
@@ -230,6 +241,19 @@ impl<Metadata: for<'a> Deserialize<'a>> Manifest<Metadata> {
                 self.bench = self.autoset("benches", &fs);
             }
         }
+        if let Some(ref mut package) = self.package {
+            if matches!(package.build, None | Some(OptionalFile::Flag(true))) {
+                if fs.file_names_in(".").map_or(false, |dir| dir.contains("build.rs")) {
+                    package.build = Some(OptionalFile::Path("build.rs".into()));
+                }
+            }
+            if matches!(package.readme, OptionalFile::Flag(true)) {
+                let files = fs.file_names_in(".").ok();
+                if let Some(name) = files.as_ref().and_then(|dir| dir.get("README.md").or_else(|| dir.get("README.txt")).or_else(|| dir.get("README"))) {
+                    package.build = Some(OptionalFile::Path((**name).to_owned()));
+                }
+            }
+        }
         Ok(())
     }
 
@@ -297,6 +321,7 @@ pub struct Profile {
     pub panic: Option<String>,
     pub incremental: Option<bool>,
     pub overflow_checks: Option<bool>,
+    pub strip: Option<bool>,
 
     /// profile overrides
     #[serde(default, serialize_with = "toml::ser::tables_last")]
@@ -411,6 +436,7 @@ pub enum Dependency {
 }
 
 impl Dependency {
+    #[inline]
     pub fn detail(&self) -> Option<&DependencyDetail> {
         match *self {
             Dependency::Simple(_) => None,
@@ -418,6 +444,7 @@ impl Dependency {
         }
     }
 
+    #[inline]
     pub fn req(&self) -> &str {
         match *self {
             Dependency::Simple(ref v) => v,
@@ -425,6 +452,7 @@ impl Dependency {
         }
     }
 
+    #[inline]
     pub fn req_features(&self) -> &[String] {
         match *self {
             Dependency::Simple(_) => &[],
@@ -432,12 +460,14 @@ impl Dependency {
         }
     }
 
+    #[inline]
     pub fn optional(&self) -> bool {
         self.detail().map_or(false, |d| d.optional)
     }
 
     // `Some` if it overrides the package name.
     // If `None`, use the dependency name as the package name.
+    #[inline]
     pub fn package(&self) -> Option<&str> {
         match *self {
             Dependency::Simple(_) => None,
@@ -446,6 +476,7 @@ impl Dependency {
     }
 
     // Git URL of this dependency, if any
+    #[inline]
     pub fn git(&self) -> Option<&str> {
         self.detail().and_then(|d| d.git.as_deref())
     }
@@ -492,14 +523,19 @@ pub struct DependencyDetail {
 /// to parse into something more useful than a generic toml `Value`
 #[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
 #[serde(rename_all = "kebab-case")]
+#[non_exhaustive]
 pub struct Package<Metadata = Value> {
     /// Careful: some names are uppercase
     pub name: String,
     #[serde(default)]
     pub edition: Edition,
+    /// MSRV 1.x (beware: does not require semver formatting)
+    #[serde(default)]
+    pub rust_version: Option<String>,
     /// e.g. "1.9.0"
     pub version: String,
-    pub build: Option<Value>,
+    #[serde(default)]
+    pub build: Option<OptionalFile>,
     pub workspace: Option<String>,
     #[serde(default)]
     /// e.g. ["Author <e@mail>", "etc"] Deprecated.
@@ -513,8 +549,8 @@ pub struct Package<Metadata = Value> {
     pub documentation: Option<String>,
     /// This points to a file under the package root (relative to this `Cargo.toml`).
     /// implied if README.md, README.txt or README exists.
-    #[serde(default, deserialize_with = "readme_parser")]
-    pub readme: Option<String>,
+    #[serde(default, skip_serializing_if = "OptionalFile::is_default")]
+    pub readme: OptionalFile,
     #[serde(default, skip_serializing_if = "Vec::is_empty")]
     pub keywords: Vec<String>,
     #[serde(default, skip_serializing_if = "Vec::is_empty")]
@@ -546,6 +582,43 @@ pub struct Package<Metadata = Value> {
     pub metadata: Option<Metadata>,
 }
 
+/// Readme of build.rs path
+#[derive(Debug, Clone, Eq, PartialEq, Serialize, Deserialize)]
+#[serde(untagged)]
+pub enum OptionalFile {
+    /// Opt-in to default, or explicit opt-out
+    Flag(bool),
+    /// Explicit path
+    Path(String),
+}
+
+impl Default for OptionalFile {
+    #[inline]
+    fn default() -> Self {
+        Self::Flag(true)
+    }
+}
+
+impl OptionalFile {
+    #[inline]
+    fn is_default(&self) -> bool {
+        matches!(self, Self::Flag(flag) if *flag)
+    }
+
+    #[inline]
+    pub fn as_ref(&self) -> Option<&str> {
+        match self {
+            Self::Path(p) => Some(p),
+            _ => None,
+        }
+    }
+
+    #[inline]
+    pub fn is_some(&self) -> bool {
+        matches!(self, Self::Flag(true) | Self::Path(_))
+    }
+}
+
 #[derive(Debug, Clone, Eq, PartialEq, Serialize, Deserialize)]
 #[serde(untagged)]
 pub enum Publish {
@@ -560,12 +633,14 @@ impl Publish {
 }
 
 impl Default for Publish {
+    #[inline]
     fn default() -> Self {
         Publish::Flag(true)
     }
 }
 
 impl PartialEq<Publish> for bool {
+    #[inline]
     fn eq(&self, p: &Publish) -> bool {
         match *p {
             Publish::Flag(flag) => flag == *self,
@@ -575,6 +650,7 @@ impl PartialEq<Publish> for bool {
 }
 
 impl PartialEq<bool> for Publish {
+    #[inline]
     fn eq(&self, b: &bool) -> bool {
         b.eq(self)
     }
@@ -692,14 +768,14 @@ impl Default for MaintenanceStatus {
     }
 }
 
-#[derive(Debug, PartialEq, Eq, Copy, Clone, Hash, Serialize, Deserialize)]
+#[derive(Debug, PartialEq, Eq, Ord, PartialOrd, Copy, Clone, Hash, Serialize, Deserialize)]
 pub enum Edition {
     #[serde(rename = "2015")]
-    E2015,
+    E2015 = 2015,
     #[serde(rename = "2018")]
-    E2018,
+    E2018 = 2018,
     #[serde(rename = "2021")]
-    E2021,
+    E2021 = 2021,
 }
 
 impl Default for Edition {
@@ -708,12 +784,25 @@ impl Default for Edition {
     }
 }
 
-#[derive(Debug, PartialEq, Eq, Copy, Clone, Hash, Serialize, Deserialize)]
+#[derive(Debug, PartialEq, Eq, Ord, PartialOrd, Copy, Clone, Hash, Serialize, Deserialize)]
 pub enum Resolver {
     #[serde(rename = "1")]
-    V1,
+    V1 = 1,
     #[serde(rename = "2")]
-    V2,
+    V2 = 2,
+}
+
+impl Display for Resolver {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        write!(
+            f,
+            "{}",
+            match self {
+                Resolver::V1 => "1",
+                Resolver::V2 => "2",
+            }
+        )
+    }
 }
 
 impl Default for Resolver {
@@ -721,43 +810,3 @@ impl Default for Resolver {
         Self::V1
     }
 }
-
-fn readme_parser<'de, D>(deser: D) -> Result<Option<String>, D::Error>
-where
-    D: Deserializer<'de>,
-{
-    use serde::de;
-    use serde::de::Visitor;
-    use std::fmt;
-
-    struct ReadmeVisitor;
-    impl<'de> Visitor<'de> for ReadmeVisitor {
-        type Value = Option<String>;
-
-        fn expecting(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
-            write!(formatter, "readme property must be a stirng or bool")
-        }
-
-        fn visit_none<E: de::Error>(self) -> Result<Self::Value, E> {
-            Ok(None)
-        }
-
-        fn visit_unit<E: de::Error>(self) -> Result<Self::Value, E> {
-            Ok(None)
-        }
-
-        fn visit_bool<E: de::Error>(self, _: bool) -> Result<Self::Value, E> {
-            Ok(None) // unfortunately the return type can't capture the difference between none and false
-        }
-
-        fn visit_str<E: de::Error>(self, val: &str) -> Result<Self::Value, E> {
-            Ok(Some(val.to_string()))
-        }
-
-        fn visit_string<E: de::Error>(self, val: String) -> Result<Self::Value, E> {
-            Ok(Some(val))
-        }
-    }
-
-    deser.deserialize_any(ReadmeVisitor)
-}
